

-- Animations System script
-- v1.2 (15th November 2023)
-- changelog from previous version (v1.1):
-- 	- Animations targeting vanilla model parts now properly layer on top of existing rotations applied by the engine, instead of overwriting them.
--		- This was a pain to implement. You're welcome!

-- This script implements a system to create and handle script-driven animations.
-- To use: 
--		From another script, create a new animation by calling DefineAnimation().
--		The function arguments that DefineAnimation() takes are as follows:
--			1) A table of one or more model parts that the animation will affect.
--				These are models.* paths - so to target your model's head, for example, it'd look something like "models.MyCoolModel.Head". "models.[filename].[part path within file]"
--				Lua's table formatting is {item1, item2, item3} etc.
--				Vanilla model parts can also be animated.
--			2) A function that controls when the animation plays.
--				This function is called every tick, and the animation will play when it returns true.
--			3) A function that drives the animation.
--				This function is given the position and rotation vector3s as arguments, and should return new values for both in that order, too.
--				The animation system handles interpolation (animation smoothing) and delta time (accounting for framerate) for you - don't worry about it.
--
--		For example:
--			DefineAnimation(
--				{models.YOURMODELNAME.LeftArm, models.YOURMODELNAME.RightArm},			-- {Table} of model parts. Don't forget the {curly braces}!
--				function() return true end,												-- When-to-play function. Don't forget to return true/false!
--				function(pos, rot) return pos, rot + vec(0,5,0) end						-- Driver function. Don't forget your arguments and returns!
--			)
--		This example function will rotate both arms of YOURMODELNAME by five degrees around the Y axis, every tick.
--		Because it modifies and returns the values given in instead of making entirely new ones, it will layer on top of other animations that come before it.
--		
--		The Figura documentation really helps! [https://applejuiceyy.github.io/figs/latest/]
--		You might also find a graphing calculator useful. There are plenty online.
--

-- In more detail:
--		- The user defines an Animation, specifying what model parts it affects, when it's active, and the driving function.
--		- The script keeps track of what model parts are affected, and makes proxy variables for their positions and rotations.
--		- Each tick, the script runs through every Animation in order, calling the driving functions with the target proxies.
--		- Each render, the script interpolates the real positions of model parts toward the target proxies, taking into account frametime.

-- Modify however you want, just give me credit if you release it.
-- written by pont
-- visit pont.cool ;3


-- Make sure you have 'pont_playerstate.lua' as well.
require("pont_playerstate")
-- This isn't a dependency, but it does make using the animationsystem much easier.
-- Feel free to comment this out if you don't want to use it. Are you rolling your own? We understand that mindset well - good luck!

-- #####################################################################################################
-- Variable Declarations
-- #####################################################################################################

tAllAnimations = {}		-- Table containing every animation the user makes with this script. Ideally populated through DefineAnimation().

-- Internal and config
local ktPartTargetValues = {} 	-- Keyed table containing part references, and target values for position and rotation to be interpolated towards.
local InterpolationValue = 0.3 	-- Config: Controls how fast animations interpolate to their target positions, where 1 is instant snapping and 0 is no movement at all.

-- #####################################################################################################
-- Script Proper
-- #####################################################################################################

-- Every tick (update function)
-- Bound to .tick within Figura's Events API
function events.tick()
	-- If there aren't any animations, return immediately.
	if #tAllAnimations == 0 then return end
	
	EnactAnimations()
end

-- Before every avatar render
-- Bound to .render within Figura's Events API
function events.render()
	
	InterpolateAll(client.getFrameTime())
end


-- Makes a new animation 'struct', and adds it to tAllAnimations.
-- This function exists to standardise formatting and catch errors.
function DefineAnimation (in_TargetParts, in_ActiveWhenFunction, in_AnimationFunction)

	-- Instantiate struct
	local NewAnimation = {
		TargetParts = {},			-- Table of model.* parts to affected by this animation
		ActiveWhenFunction,			-- The animation will play when this function returns true
		AnimationFunction			-- Function that defines the animation. Run for each part in TargetParts.
		-- Part's relative position is passed in the first argument.
		-- Part's relative rotation is passed in the second argument.
	}
	
	-- Initialise struct
	-- TargetParts
	-- Bad input catch
	if in_TargetParts == nil then error("DefineAnimation(): Called but not given any model parts.") end
	-- For every given Target Part...
	for i_TargetKey, i_TargetPart in pairs(in_TargetParts) do -- pairs(table) is some Lua quirk.
		-- Bad input catch 
		if i_TargetPart == nil then error("DefineAnimation(): '" .. i_TargetPart .. "' can't be found in model file.") end
		-- Push to table
		table.insert(NewAnimation.TargetParts, i_TargetPart);
		
		-- Add part to keytable
		-- Partname is the key (as to avoid duplicate entries), value is a table of two vector3s (for relative position, and relative rotation).
		ktPartTargetValues[i_TargetPart] = {TargetPos = vec(0,0,0), TargetRot = vec(0,0,0)}
	end
	-- ActiveWhenFunction
	NewAnimation.ActiveWhenFunction = in_ActiveWhenFunction
	-- AnimationFunction
	NewAnimation.AnimationFunction = in_AnimationFunction
	-- Doesn't look like checking these functions' arguments and returns is possible - can't throw user-friendly errors.
	
	-- Add this new animation to tAllAnimations.
	table.insert(tAllAnimations, NewAnimation)
	
	-- Return animation struct, in case you need it for something.
	return NewAnimation
end

-- Enacts animations' functions on target values.
function EnactAnimations()

	-- For every animation...
	for i_AnimKey, i_Animation in pairs(tAllAnimations) do
	
		-- For every target part of that animation...
		for i_TargetKey, i_TargetPart in pairs(i_Animation.TargetParts) do
		
			if i_Animation.ActiveWhenFunction() then
				-- Enact animation function on data in target values table
				-- This is a big line! Let's break it down.
				-- We're running this animation's AnimationFunction on each of the animation's target parts, passing in the current target part's proxy data and writing out to them, too.
				-- So:
				ktPartTargetValues[i_TargetPart].TargetPos, ktPartTargetValues[i_TargetPart].TargetRot = 	-- Writing to this part's target position and rotation...
					i_Animation.AnimationFunction(															-- the result of the AnimationFunction.
						ktPartTargetValues[i_TargetPart].TargetPos,											-- Giving the AnimationFunction the position proxy variable
						ktPartTargetValues[i_TargetPart].TargetRot											-- Giving the AnimationFunction the rotation proxy variable
					)																						-- Closing the function arguments
				
			end
			-- If ActiveWhen function end
		end
		-- For each target part end
	end
	-- For each animation end
end
-- EnactAnimations() end


-- Interpolates actual mesh parts toward target values.
function InterpolateAll(in_FrameTime)

	-- Declare temporary variables outside of loop for performance
	-- I don't know how clever Lua's interpreter is about declarations within loops - playing it safe.
	local temp_InterpDTime = InterpolationValue * in_FrameTime
	local temp_PosDelta			-- Position change, working variable
	local temp_RotDelta			-- Rotation change, working variable
	local temp_RealPartPos		-- Real position value taken from part, separate from animsys proxy values
	local temp_RealPartRot		-- Real rotation value taken from part, separate from animsys proxy values
	
	-- For every targeted part...
	for i_TargetPart, i_TargetValues in pairs(ktPartTargetValues) do
	
		-- Check what type of model part i_TargetPart is
		-- Custom model parts and Vanillamodel parts have to be handled separately.
		-- getName() only exists on custom model parts.
		if i_TargetPart.getName ~= nil then
			-- This is a CUSTOM MODEL PART.
			
			--log("CUSTOM", i_TargetPart, "getName?", i_TargetPart.getName) -- debug (LAGGY)
			
			-- Get real values from part
			temp_RealPartPos = i_TargetPart:getPos()
			temp_RealPartRot = i_TargetPart:getRot()
			
			-- Calc deltas
			-- Position
			temp_PosDelta = i_TargetValues.TargetPos - temp_RealPartPos
			-- Rotation
			temp_RotDelta = i_TargetValues.TargetRot - temp_RealPartRot
			
			-- Affect by interp multiplier
			temp_PosDelta = temp_PosDelta * temp_InterpDTime
			temp_RotDelta = temp_RotDelta * temp_InterpDTime
			
			-- Add to existing values
			temp_PosDelta = temp_PosDelta + temp_RealPartPos
			temp_RotDelta = temp_RotDelta + temp_RealPartRot
			
			-- Write new values to 
			i_TargetPart:setPos(temp_PosDelta)
			i_TargetPart:setRot(temp_RotDelta)
		
		else
			-- This is a VANILLA MODEL PART.
			-- Vanillamodel parts need to be handled a little differently.
			
			--log("VANILLA", i_TargetPart, "getName?", i_TargetPart.getName) -- debug (LAGGY)
			
			-- Get real values from part
			temp_RealPartPos = (i_TargetPart:getPos() or vec(0,0,0))
			temp_RealPartRot = (i_TargetPart:getOffsetRot() or vec(0,0,0))
			-- Safety netting - vanillamodel parts' values are nil at first, and rotation offset values use a different function name.
			
			-- Calc deltas
			-- Position
			temp_PosDelta = i_TargetValues.TargetPos - temp_RealPartPos
			-- Rotation
			temp_RotDelta = i_TargetValues.TargetRot - temp_RealPartRot
			
			-- Affect by interp multiplier
			temp_PosDelta = temp_PosDelta * temp_InterpDTime
			temp_RotDelta = temp_RotDelta * temp_InterpDTime
			
			-- Add to existing values
			temp_PosDelta = temp_PosDelta + temp_RealPartPos
			temp_RotDelta = temp_RotDelta + temp_RealPartRot
			
			-- Write new values to 
			i_TargetPart:setPos(temp_PosDelta)
			i_TargetPart:setOffsetRot(temp_RotDelta)
		end
	
	
	end
	-- For each targeted part end
	
	
	-- DEBUG READOUT
	--("INTERP", "i"..InterpolationValue, "f"..in_FrameTime, "\nif"..temp_InterpDTime)
	
end





-- call us the Data Structurers the way we . are doing that