

-- Player State script
-- v1.4 (22nd November 2023)
-- changelog from previous version (v1.3):
--	- Fixed a player standing slightly within water, but with their origin outside of it, counting as InRain.
--		This edge case now counts as Wading.
--	- Fixed a player bouncing atop a bubble column sometimes, briefly, counting as InRain or Wading.
--	- Added a commented-out version of the sneaking check for the logical sneaking state instead of the visual one, in case you need that.
--		Default script behaviour remains unchanged here.

-- This script reads/calcs some basic information about the player entity's state and bundles it into a table for ease-of-use.
-- eg. velocity, angular velocity, sprinting, in water, etc.
-- To use: read data from 'tPlayerEntityInfo' (defined below).
-- This isn't all-inclusive - see Figura documentation [https://applejuiceyy.github.io/figs/latest] if you need to check more specific things.

-- Prefer an eye on the data? There's a big chunk of debug log()s commented out near the end of the file.
-- Debug for state enums is nestled into the logic - because they're just objects, the interpreter can't see their names and prints a memory address.

-- Modify however you want, just give me credit if you release it.
-- written by pont
-- visit pont.cool ;3


-- #####################################################################################################
-- Variable Declarations
-- #####################################################################################################

-- Lua doesn't really do enumerables, but you can fudge it with blank table members like this. Empty objects do evaluate correctly.
-- 'Enum' describing player movement states - eg standing, walking, sprinting, etc.
ePlayerMoveState = {
	Standing = {},		-- Not moving
	Walking = {},		-- Walking around
	Sprinting = {},		-- Sprinting/running around
	Sneaking = {},		-- Sneaking/crouching
	Prone = {},			-- Crawling/diving (lying on front) - check tPlayerEntityInfo.WaterState as well to determine which.
	Climbing = {},		-- On a ladder/vine/etc
	Sitting = {},		-- On a vehicle/mount
	Sleeping = {},		-- In a bed (lying on back)
	Dead = {},			-- Dead and hasn't clicked respawn yet.
	--
	Falling = {},		-- Airborne for a little while, descending
	FallingUp = {},		-- Airborne for a little while, ascending
	CreativeFly = {},	-- Creative flight
	ElytraFly = {},		-- Elytra flight
	Riptide = {}		-- Riptide spin
}
-- 'Enum' describing player water states - because you can do a lot of things while in water, it makes more sense to check this separately.
-- eg. if a player's movement state is Prone, and their water state is Submerged, they're diving/sprint-swimming/whatever you call it.
ePlayerWaterState = {
	Dry = {},			-- Completely dry
	InRain = {},		-- Standing in rain
	InBubbles = {},		-- In a bubble column
	Wading = {},		-- Partially in water
	Submerged = {}		-- Completely underwater
}

-- Table of data about player's current state.
-- This is the part you should be reading from other scripts. Don't write to here, it'll be overwritten.
tPlayerEntityInfo = {
	Position = vec(0,0,0), 		-- Current position this tick
	Angle = vec(0,0), 			-- Current head rotation this tick
	BodyYaw = 0,				-- Current body rotation this tick
	--
	Velocity = vec(0,0,0), 		-- Position delta from previous tick
	AngularVelocity = vec(0,0),	-- Head rotation delta from previous tick
	BodyAngularVelocity = 0,	-- Body rotation delta from previous tick
	--
	MoveState = ePlayerMoveState.Standing,	-- Movement state
	WaterState = ePlayerWaterState.Dry,		-- Water state
	--
	Time = 0	-- Time, in ticks, this script has been running. Useful for script-driven animations.
}
-- Internal and config
local LastTickAngle = vec(0,0) 				-- Last tick's player head rotations
local LastTickBodyYaw = 0					-- Last tick's player body rotation
local FallingThreshold_Time = 15 			-- Config: Time, in ticks, for a player to be airborne to count as Falling.
	-- Jumps along flat ground take about 11-12 ticks.
local FallingThreshold_Velocity = 0.4		-- Config: Velocity threshold for a player to count as Falling, regardless of duration.
	-- Peak ascent rate of a player jump is about ~+y0.375.
local GroundedCheckOffset = vec(0,-0.1,0) 	-- Config: Offset from player position to check for a solid block beneath the player.
	-- Player entity position is measured from the entity origin, at the feet.
local TimeAirborne = 0						-- Ticks the player has been off the ground for
local WadingCheckOffset = vec(0,0.7,0)		-- Config: Offset from player position to check for a waterlogged block at player's waist height.
local BubblesCheckOffset = vec(0,-3,0)		-- Config: Offset from player position to check for bubble columns beneath the player.
local StandingStillThreshold = 0.05			-- Config: Horizontal velocity tolerance, below which a player is considered to be standing still.
	-- Player walking velocity is about ~0.215. Sprinting is about ~0.280. Sneaking is about ~0.065.


-- #####################################################################################################
-- Script Proper
-- #####################################################################################################

-- Every tick (update function)
-- Bound to .tick within Figura's Events API
function events.tick()
	UpdateVars()
end

-- Update player state information
function UpdateVars()
	-- Store last tick's head angle before reading new values
	LastTickAngle = tPlayerEntityInfo.Angle
	-- Store last tick's body angle beffore reading new value
	LastTickBodyYaw = tPlayerEntityInfo.BodyYaw
	
	-- VAR READ/CALC
	-- Read player position
	tPlayerEntityInfo.Position = player:getPos()
	-- Read player head angle
	-- Reads camera - actual playermodel head yaw is very slightly smoothed. You might want to use vanilla_model.HEAD:getOriginRot() for remaking head visuals without parenting.
	tPlayerEntityInfo.Angle = player:getRot()
	-- Read player body yaw
	tPlayerEntityInfo.BodyYaw = player:getBodyYaw()
	-- Read player velocity
	tPlayerEntityInfo.Velocity = player:getVelocity()
	-- Calc player head angular velocity (angle delta)
	tPlayerEntityInfo.AngularVelocity = LastTickAngle - tPlayerEntityInfo.Angle
	-- Calc player body yaw angular velocity (yaw delta)
	tPlayerEntityInfo.BodyAngularVelocity = LastTickBodyYaw - tPlayerEntityInfo.BodyYaw
	
	-- Zero out velocities on first tick
	if tPlayerEntityInfo.Time == 0 then
		tPlayerEntityInfo.Velocity = vec(0,0,0)
		tPlayerEntityInfo.AngularVelocity = vec(0,0)
		tPlayerEntityInfo.BodyAngularVelocity = 0
	end
	
	-- PLAYER STATE CHECKS
	-- Airborne
	-- Is the player grounded?
	-- 		- Are they considered on the ground by the engine? This is buggy before 1.20.
	-- 		- Is the block beneath them considered solid? This doesn't catch things like glass.
	-- 		- Are they in a vehicle?
	if player:isOnGround() or world.getBlockState(tPlayerEntityInfo.Position + GroundedCheckOffset):isSolidBlock() or player:getVehicle() ~= nil then
		-- Yes, player is standing on solid ground, or is in a vehicle
		-- Set ticker to 0
		TimeAirborne = 0
		-- Ticker is also set to 0 by some WaterState checks.
	else
		-- No, player is airborne
		-- Increment ticker
		TimeAirborne = TimeAirborne + 1 -- Lua doesn't have increment shorthand?!
	end
	
	
	-- Movement
	-- Is the player climbing a ladder/vine/etc?
	if player:isClimbing() then
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Climbing
		--log("CLIMBING") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	-- Is the player prone? 'Swimming' pose can be both in water, and in one-block-tall dry spaces. Check WaterState.
	if player:getPose() == "SWIMMING" then
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Prone
		--log("PRONE") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	-- Is the player sneaking?
	if player:getPose() == "CROUCHING" then 	-- This line checks for sneak visuals only
	--if player:isSneaking() then 				-- This line checks for sneak behaviour, regardless of if you appear as sneaking or not
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Sneaking
		--log("SNEAKING") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	-- Is the player sleeping?
	if player:getPose() == "SLEEPING" then
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Sleeping
		--log("SLEEPING") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	-- Is the player in a vehicle?
	if player:getVehicle() ~= nil then
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Sitting
		--log("SITTING") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	-- Is the player elytra gliding?
	if player:isGliding() then
		tPlayerEntityInfo.MoveState = ePlayerMoveState.ElytraFly
		--log("FLYING") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	-- Is the player launching with a riptide trident?
	if player:getPose() == "SPIN_ATTACK" then
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Riptide
		--log("RIPTIDE") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	-- Is the player dead?
	if player:getHealth() == 0 then
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Dead
		--log("DEAD") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	
	-- Is the player stationary, AND on the ground?
	if tPlayerEntityInfo.Velocity:length() < StandingStillThreshold and TimeAirborne == 0 then
		-- Yes, player's standing still
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Standing
		--log("STANDING") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	
	
	-- Is the player airborne for longer than the configured threshold?
	if TimeAirborne > FallingThreshold_Time then
		-- Is the player falling faster than the configured falling threshold?
		if math.abs(tPlayerEntityInfo.Velocity.y) > math.abs(FallingThreshold_Velocity) then
			-- Yes, large Y velocity
			-- Are they going up?
			if tPlayerEntityInfo.Velocity.y > 0 then
				tPlayerEntityInfo.MoveState = ePlayerMoveState.FallingUp
				--log("FALLING UP") -- debug
				goto State_Check_Done -- Skip rest of checks
			else
				-- No, they're going down.
				tPlayerEntityInfo.MoveState = ePlayerMoveState.Falling
				--log("FALLING DOWN") -- debug
				goto State_Check_Done -- Skip rest of checks
			end
		end
		-- Is falling faster than threshold end
		
		-- Are they in Creative mode?
		if player:getGamemode() == "CREATIVE" then
			--log("CREATIVE FLIGHT") -- debug
			tPlayerEntityInfo.MoveState = ePlayerMoveState.CreativeFly
			goto State_Check_Done -- Skip rest of checks
		end
	end
	-- Is airborne end
	
	-- By this point, we've ruled everything else out.
	-- The player is moving, isn't considered falling, isn't in any other special state - they're just travelling across the ground.
	
	-- Is the player sprinting?
	if player:isSprinting() then
		tPlayerEntityInfo.MoveState = ePlayerMoveState.Sprinting
		--log("SPRINTING") -- debug
		goto State_Check_Done -- Skip rest of checks
	end
	
	-- Player is walking.
	-- There's no other possibility by this point.
	tPlayerEntityInfo.MoveState = ePlayerMoveState.Walking
	--log("WALKING") -- debug
	--goto State_Check_Done -- Skip rest of checks
	
	
	::State_Check_Done:: -- goto target
	-- Movement state checks end
	
	-- Water
	-- Is player wet?
	if player:isWet() then
		-- Is player in a bubble column?
		-- The second check exists to prevent flickers of 'in rain' when a player is bouncing atop a bubble column.
		if world.getBlockState(tPlayerEntityInfo.Position + WadingCheckOffset):getID() == "minecraft:bubble_column" or
		   world.getBlockState(tPlayerEntityInfo.Position + BubblesCheckOffset):getID() == "minecraft:bubble_column" then
			tPlayerEntityInfo.WaterState = ePlayerWaterState.InBubbles
			TimeAirborne = 0 -- Force airborne ticker to 0
			--log("IN BUBBLES") -- debug
			goto Water_Check_Done -- Skip rest of checks
		end
		-- Is player's eye level underwater?
		if player:isUnderwater() then
			tPlayerEntityInfo.WaterState = ePlayerWaterState.Submerged
			TimeAirborne = 0 -- Force airborne ticker to 0
			--log("SUBMERGED") -- debug
			goto Water_Check_Done -- Skip rest of checks
		end
		
		-- Is player wading/partially submerged?
		if player:isInWater() then
			-- Player is partially submerged
			tPlayerEntityInfo.WaterState = ePlayerWaterState.Wading
			TimeAirborne = 0 -- Force airborne ticker to 0
			--log("WADING") -- debug
			goto Water_Check_Done -- Skip rest of checks
		end
		
		-- Player is wet, but not in water. It's probably rain!
		tPlayerEntityInfo.WaterState = ePlayerWaterState.InRain
		--log("IN RAIN") -- debug
		goto Water_Check_Done -- Skip rest of checks
	else
		-- Player is dry.
		tPlayerEntityInfo.WaterState = ePlayerWaterState.Dry
		--log("DRY") -- debug
		goto Water_Check_Done -- Skip rest of checks
	end
	::Water_Check_Done::
	-- Water state checks end
	
	
	-- Increment time ticker
	tPlayerEntityInfo.Time = tPlayerEntityInfo.Time + 1
	
	
	-- DEBUG READOUT
	--log("airborne:", TimeAirborne)
	--log("pos:", tPlayerEntityInfo.Position)
	--log("vel:", tPlayerEntityInfo.Velocity)
	--log("rot:", tPlayerEntityInfo.Angle)
	--log("avel:", tPlayerEntityInfo.AngularVelocity)
	--log("bYaw:", tPlayerEntityInfo.BodyYaw, " avel: ", tPlayerEntityInfo.BodyAngularVelocity)
	--log("##################################")
	
end

-- When called by require() from another script, return the player data struct.
return tPlayerEntityInfo





-- hi can we just take a moment here 
-- only nil and false are falsy in lua.
-- i dont care about the start-at-1 indexing, i dont care about it being tables all the way down
-- but 0 evaluating as true?
-- youre insane
--
-- what the hell
--
--
-- dude
--
